; Copyright 2019-2021 Sean Porter (Glitchfinder)
;
; Permission is hereby granted, free of charge, to any person obtaining a copy
; of this software and associated documentation files (the "Software"), to deal
; in the Software without restriction, including without limitation the rights
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
; copies of the Software, and to permit persons to whom the Software is
; furnished to do so, subject to the following conditions:
;
; The above copyright notice and this permission notice shall be included in all
; copies or substantial portions of the Software.
;
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
; SOFTWARE.
;------------------------------------------------------------------------------/

Scriptname zYAWA_Patrol_RR Extends ObjectReference

zYAWA_Settlement_RR[] Property Settlements  Auto Hidden ; The ends of this patrol
Actor[]            Property Patrollers   Auto Hidden ; The patrolling actors
Float              Property Respawn      Auto Hidden ; ETA for spawn/respawn
Bool               Property Eyebot       Auto Hidden ; Is there an eyebot?
Bool               Property Robot        Auto Hidden ; Is there a robot?
Bool               Property Updated      Auto Hidden ; Prevent double-updates
Bool               Property ForceRespawn Auto Hidden ; Ignore respawn timeline?
Bool               Property Moved        Auto Hidden ; Did a settlement move?
Bool               Property RegionUpdate Auto Hidden ; Did a region update?
Bool               Property UpdateBase   Auto Hidden ; Change the ActorBase?
Int                Property Size         Auto Hidden ; Number of patrollers
Int                Property Distance     Auto Hidden ; Maximum patrol distance
Int                Property Productivity Auto Hidden ; Minimum settlement prod.

;------------------------------------------------------------------------------\
; Initialize this patrol
;------------------------------------------------------------------------------/
Function Initialize(Int one, Int two, zYAWA_Manager_RR m)
	; Build the Settlement list
	Settlements = New zYAWA_Settlement_RR[2]
	Settlements[0] = m.Settlements[one]
	Settlements[1] = m.Settlements[two]

	; Create the patroller list
	Patrollers = New Actor[6]

	; Default patrol settings
	Respawn = -1
	Updated = False
	UpdateRegion(m)
EndFunction

;------------------------------------------------------------------------------\
; Get the regional settings for this patrol
;------------------------------------------------------------------------------/
Function UpdateRegion(zYAWA_Manager_RR m)
	; Get this patrol's current regions
	zYAWA_Region_RR r1 = Settlements[0].Region
	zYAWA_Region_RR r2 = Settlements[1].Region

	; Get this patrol's size, distance, and productivity settings
	Size = Math.Max(r1.Size, r2.Size) As Int
	Distance = m.Densities[Math.Max(r1.Density, r2.Density) As Int]
	Productivity = m.Products[Math.Min(r1.Productivity, r2.Productivity) As Int]

	; If robots are enabled and either region is set to use them
	If (m.Robots && (r1.Robot || r2.Robot))
		; This patrol has a robot
		Robot = True

		; If this patrol does not currently have a robot, and already exists
		If (Patrollers[4] == None && Respawn != -1)
			; This patrol needs to respawn wth the new member
			ForceRespawn = True
		EndIf
	; If robots are disabled or neither region uses them
	Else
		; This patrol has no robots
		Robot = False

		; Remove the robot patroller, if they exist
		RemovePatroller(4)
	EndIf

	; If eyebots are enabled and either region is set to use them
	If (m.Eyebots && (r1.Eyebot || r2.Eyebot))
		; This patrol has an eyebot
		Eyebot = True

		; If this patrol does not currently have an eyebot, and already exists
		If (Patrollers[5] == None && Respawn != -1)
			; This patrol needs to respawn with the new member
			ForceRespawn = True
		EndIf
	; If eyebots are disabled or neither region uses them
	Else
		; This patrol has no eyebots
		Eyebot = False
		
		; Remove the eyebot patroller, if they exist
		RemovePatroller(5)
	EndIf

	; This patrol has up to date region settings
	RegionUpdate = False
EndFunction

;------------------------------------------------------------------------------\
; Update this patrol
;------------------------------------------------------------------------------/
Function Update(zYAWA_Manager_RR m, Bool newHome)
	; Patrols are shared by two settlements, only update them every other call
	Updated = !Updated

	; If the settlement has moved, make sure that gets updated
	If (newHome)
		Moved = True
	EndIf

	; Only update this patrol once per tick
	If (!Updated)
		Return
	EndIf

	; Increment the patrol index so the next tick can pick up where it left off
	m.IndexThisTick += 1

	; Grab this patrol's settlements
	zYAWA_Settlement_RR s1 = Settlements[0]
	zYAWA_Settlement_RR s2 = Settlements[1]

	; If this patrol has already been processed in a prior tick, do nothing
	If (m.IndexThisTick < m.IndexLastTick)
		Return
	EndIf

	; If this patrol is beyond the rate limit for this tick, do nothing
	If (m.UpdatedThisTick >= m.UpdatesPerTick && m.UpdatesPerTick > 0)
		Return
	; If this patrol is not beyond the rate limit for this tick
	Else
		; Update the rate limit data
		m.UpdatedThisTick += 1
		m.SkippedThisTick += 1
	EndIf

	; If this patrol's settlements have moved since the last update
	If (Moved)
		; Update the patrol routes and check regional settings
		SetPatrolEndpoints(m)
		UpdateRegion(m)
	EndIf

	; Does this patrol need to bypass normal respawn restrictions?
	ForceRespawn = UpdateBase

	; If the region information is outdated, update it
	If (RegionUpdate || m.Robots != Robot || m.Eyebots != Eyebot)
		UpdateRegion(m)
	EndIf

	; Remove extra patrollers if the patrol size went down
	PrunePatrol()

	; Get the patrol endpoints
	ObjectReference one = s1.Center
	ObjectReference two = s2.Center

	; If an endpoint does not exist
	If (one == None || two == None)
		; Remove the patrol and exit processing
		RemovePatrol()
		Return
	EndIf

	; If the patrol goes too far
	If (one.GetDistance(two) > Distance && Distance != -1)
		; Remove the patrol and exit processing
		RemovePatrol()
		Return
	EndIf

	; Get the patrol's target settlements
	WorkshopScript a = s1.Settlement
	WorkshopScript b = s2.Settlement

	; If a settlement no longer exists
	If (a == None || b == None)
		; Rmove the patrol and exit processing
		RemovePatrol()
		Return
	EndIf

	; If the first settlement is no longer owned by the player, or went raider
	If (!a.OwnedByPlayer || a.HasKeyword(m.Raider))
		; Clean up dead or disabled patrol members and exit processing
		CleanPatrol()
		Return
	EndIf

	; If the second settlement is no longer owned by the player, or went raider
	If (!b.OwnedByPlayer || b.HasKeyword(m.Raider))
		; Clean up dead or disabled patrol members and exit processing
		CleanPatrol()
		Return
	EndIf

	; Initialize the production value for a settlement
	Int value = 0

	; Get the settlement's current production values
	value += (a.GetValue(m.WPS.WorkshopRatings[m.Food].resourceValue) as Int)
	value += (a.GetValue(m.WPS.WorkshopRatings[m.Water].resourceValue) as Int)
	value += (a.GetValue(m.WPS.WorkshopRatings[m.Safety].resourceValue) as Int)

	; If the settlement is not producing enough
	If (value < Productivity)
		; Clean up dead or disabled patrol members and exit processing
		CleanPatrol()
		Return
	EndIf

	; Initialize the production value for a settlement
	value = 0

	; Get the settlement's current production values
	value += (b.GetValue(m.WPS.WorkshopRatings[m.Food].resourceValue) as Int)
	value += (b.GetValue(m.WPS.WorkshopRatings[m.Water].resourceValue) as Int)
	value += (b.GetValue(m.WPS.WorkshopRatings[m.Safety].resourceValue) as Int)

	; If the settlement is not producing enough
	If (value < Productivity)
		; Clean up dead or disabled patrol members and exit processing
		CleanPatrol()
		Return
	EndIf

	; If insufficient time has passed, and a respawn is not necessary
	If (Utility.GetCurrentGameTime() < Respawn && !ForceRespawn)
		; Exit processing, since the only thing left is respawning the patrol
		Return
	EndIf

	; If the first spawnpoint is loaded
	If (one.Is3DLoaded())
		; If the second spawnpoint is loaded as well, exit processing
		If (two.Is3DLoaded())
			Return
		EndIf
		
		; Swap the spawnpoints so patrols will spawn offscreen
		ObjectReference three = one
		one = two
		two = three
	EndIf

	; Set the respawn time to a week after the current ingame time
	Respawn = Utility.GetCurrentGameTime() + 7
	
	; Get the base actor for settlers
	ActorBase toSpawn = m.PatrollerS

	; If patrols aren't supposed to use settlers, switch to minutemen
	If (!m.UseSettlerBase)
		toSpawn = m.PatrollerM
	EndIf

	; Start with the first patroller
	Int c = 0

	; Go through the full number of patrollers for this patrol
	While (c < Size)
		; Spawn this patroller and set their mortality
		SpawnPatroller(c, m, one, toSpawn)
		SetMortality(c, m)
		
		; Move to the next patroller
		c += 1
	EndWhile

	; If this patrol has a robot, spawn them and set their mortality
	If (Robot)
		SpawnPatroller(4, m, one, m.Robot)
		SetMortality(4, m)
	EndIf

	; If this patrol has an eyebot, spawn them and set their mortality
	If (Eyebot)
		SpawnPatroller(5, m, one, m.Eyebot)
		SetMortality(5, m)
	EndIf

	; Link up the spawned patrollers so they will patrol as a group
	LinkPatrol(0, m.Follow1)
	LinkPatrol(1, m.Follow2)
	LinkPatrol(2, m.Follow3)
	LinkPatrol(3, m.Follow4)
	LinkPatrol(4, m.Follow5)

	; Mark the ActorBase as no longer in need of updating
	UpdateBase = False
EndFunction

;------------------------------------------------------------------------------\
; Remove excess patrollers from this patrol
;------------------------------------------------------------------------------/
Function PrunePatrol()
	; Start at the first unused patroller slot
	Int c = Size

	; Go through all unused patroller slots
	While (c < 4)
		; Remove this patroller and move on to the next one
		RemovePatroller(c)
		c += 1
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Remove dead or disabled actors from this patrol
;------------------------------------------------------------------------------/
Function CleanPatrol()
	; Check every slot in the patrol
	CleanPatroller(0)
	CleanPatroller(1)
	CleanPatroller(2)
	CleanPatroller(3)
	CleanPatroller(4)
	CleanPatroller(5)
EndFunction

;------------------------------------------------------------------------------\
; Remove a dead or disabled actor from this patrol
;------------------------------------------------------------------------------/
Function CleanPatroller(Int id)
	; If this patroller does not exist, do nothing
	If (Patrollers[id] == None)
		Return
	EndIf

	; If this patroller is alive or currently loaded, do nothing
	If (!Patrollers[id].IsDead() || Patrollers[id].Is3DLoaded())
		Return
	EndIf

	; Disable this patroller, delete them, and clear their slot
	Patrollers[id].Disable()
	Patrollers[id].Delete()
	Patrollers[id] = None
EndFunction

;------------------------------------------------------------------------------\
; Remove all patrollers from this patrol
;------------------------------------------------------------------------------/
Function RemovePatrol()
	; Check every slot in the patrol
	RemovePatroller(0)
	RemovePatroller(1)
	RemovePatroller(2)
	RemovePatroller(3)
	RemovePatroller(4)
	RemovePatroller(5)

	; Reset the respawn time, in case settings change
	Respawn = -1
EndFunction

;------------------------------------------------------------------------------\
; Remove a patroller from this patrol
;------------------------------------------------------------------------------/
Function RemovePatroller(Int id)
	; If this patroller does not exist, do nothing
	If (Patrollers[id] == None)
		Return
	EndIf

	; Disable this patroller, delete them, and clear their slot
	Patrollers[id].Disable()
	Patrollers[id].Delete()
	Patrollers[id] = None
EndFunction

;------------------------------------------------------------------------------\
; Move this patrol's route
;------------------------------------------------------------------------------/
Function SetPatrolEndpoints(zYAWA_Manager_RR m)
	; Check every slot in this patrol
	SetPatrollerEndpoints(m, 0)
	SetPatrollerEndpoints(m, 1)
	SetPatrollerEndpoints(m, 2)
	SetPatrollerEndpoints(m, 3)
	SetPatrollerEndpoints(m, 4)
	SetPatrollerEndpoints(m, 5)

	; This patrol's endpoints are now up to date
	Moved = False
EndFunction

;------------------------------------------------------------------------------\
; Move an individual patroller's route
;------------------------------------------------------------------------------/
Function SetPatrollerEndpoints(zYAWA_Manager_RR m, Int id)
	; If this patroller does not exist, do nothing
	If (Patrollers[id] == None)
		Return
	EndIf

	; Update this patroller's route
	Patrollers[id].SetLinkedRef(Settlements[0].Center, m.DestStart)
	Patrollers[id].SetLinkedRef(Settlements[1].Center, m.DestEnd)
	Patrollers[id].SetLinkedRef(Settlements[0].Entrance, m.DoorStart)
	Patrollers[id].SetLinkedRef(Settlements[1].Entrance, m.DoorEnd)
EndFunction

;------------------------------------------------------------------------------\
; Validate Patroller
;------------------------------------------------------------------------------/
Bool Function ValidPatroller(Int id)
	; If the patroller does not exist, they are invalid
	If (Patrollers[id] == None)
		Return False
	; If the patroller is dead, they are invalid
	ElseIf (Patrollers[id].IsDead())
		Return False
	; If the patroller is disabled or deleted, they are invalid
	ElseIf (Patrollers[id].IsDisabled() || Patrollers[id].IsDeleted())
		Return False
	EndIf

	; This patroller is valid
	Return True
EndFunction

;------------------------------------------------------------------------------\
; Spawn Patroller
;------------------------------------------------------------------------------/
Function SpawnPatroller(Int id, zYAWA_Manager_RR m, ObjectReference c, ActorBase b)
	; If this patroller does not exist, is dead, is disabled, or is deleted
	If (!ValidPatroller(id))
		; Spawn a new patroller in their place
		Patrollers[id] = c.PlaceActorAtMe(b)
	; If a valid patroller already has this slot
	Else
		; Disable and delete them
		Patrollers[id].Disable()
		Patrollers[id].Delete()
		; Spawn a new patroller
		Patrollers[id] = c.PlaceActorAtMe(b)
	EndIf

	; Set the new patroller's route
	Patrollers[id].SetLinkedRef(Settlements[0].Center, m.DestStart)
	Patrollers[id].SetLinkedRef(Settlements[1].Center, m.DestEnd)
	Patrollers[id].SetLinkedRef(Settlements[0].Entrance, m.DoorStart)
	Patrollers[id].SetLinkedRef(Settlements[1].Entrance, m.DoorEnd)
EndFunction

;------------------------------------------------------------------------------\
; Link this patroller into their patrol
;------------------------------------------------------------------------------/
Function LinkPatrol(Int id, Keyword type)
	; Start at the back of the patrol
	Int c = 5

	; Check every patroller except the first, who follows no one
	While (c > 0)
		; If the new patroller is behind the one being checked, we're done
		If (id >= c)
			Return
		EndIf

		; If the patroller being checked exists
		If (Patrollers[c] != None)
			; Add this patroller as a follow target
			Patrollers[c].SetLinkedRef(Patrollers[id], type)
		EndIf

		; Move to the next patroller in line
		c -= 1
	EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Update this patrol's mortality
;------------------------------------------------------------------------------/
Function UpdateMortality(zYAWA_Manager_RR m)
	; Set a patroller index
		Int ci = 0

		; Go through the list of patrollers
		While (ci < 6)
			; Update the patroller's mortality setting
			SetMortality(ci, m)
			; Move to the next patroller
			ci += 1
		EndWhile
EndFunction

;------------------------------------------------------------------------------\
; Set a patroller's mortality
;------------------------------------------------------------------------------/
Function SetMortality(Int id, zYAWA_Manager_RR m)
	; If this patroller is dead, disabled, or doesn't exist, exit processing
	If (!ValidPatroller(id))
		Return
	EndIf
	
	; If patrollers are protected, and not the player's enemy
	If (m.Mortality == 1 && m.Minutemen.IsPlayerEnemy() != True)
		; Set the patroller as protected
		Patrollers[id].SetProtected(True)
		Return
	EndIf
	
	; If patrollers are essential, and not the player's enemy
	If (m.Mortality == 2 && m.Minutemen.IsPlayerEnemy() != True)
		; Set the patroller as essential
		Patrollers[id].SetEssential(True)
		Return
	EndIf
	
	; Set this patroller as mortal
	Patrollers[id].SetProtected(False)
	Patrollers[id].SetEssential(False)
EndFunction
